@@ -1,7 +1,7 @@
 #ifndef __PLINK2_BGZF_H__
 #define __PLINK2_BGZF_H__
 
-// This library is part of PLINK 2.00, copyright (C) 2005-2023 Shaun Purcell,
+// This library is part of PLINK 2.00, copyright (C) 2005-2024 Shaun Purcell,
 // Christopher Chang.
 //
 // This library is free software: you can redistribute it and/or modify it
@@ -35,11 +35,10 @@
 
 #include "plink2_string.h"
 #include "plink2_thread.h"
-#ifdef TRY_SYSTEM_LIBDEFLATE
+#ifdef IGNORE_BUNDLED_LIBDEFLATE
 #  include <libdeflate.h>
 #else
-// Usually requires ../libdeflate to be in include path.
-#  include "libdeflate.h"
+#  include "../libdeflate/libdeflate.h"
 #endif
 
 #ifdef __cplusplus
